(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright Inria 2021
*)

module AnalysisKind = struct
  type t =
    | IndependentAttribute
  [@@deriving enum]

  let to_string = function
    | IndependentAttribute -> "independent attribute"
end

module Options = struct
  type t = {
    run_internal_tests: bool;
    learn_from_patterns: bool;
    strict_errors: bool;
    unroll_fixpoints: int;
    delay_widening: int;
    analysis_kind: AnalysisKind.t;
  }

  module type S = sig
    val extensional_eq : bool
    val extensional_bot: bool
    val learn_from_patterns: bool
    val strict_errors: bool
    val unroll_fixpoints: int
    val delay_widening: int
  end

  let get_module options =
    let module Options = struct
      let extensional_eq = false
      let extensional_bot = false
      let learn_from_patterns = options.learn_from_patterns
      let strict_errors = options.strict_errors
      let unroll_fixpoints = options.unroll_fixpoints
      let delay_widening = options.delay_widening
    end in
    (module Options : S)

  module type ANALYZER = sig
    module Env: sig
      type 'a t
      val empty : 'a t
    end
    module Rel: sig
      type t
    end

    type env = Rel.t Env.t

    val interpret_program:
      env ->
      Ast.Inner.program ->
      (Ast.Var.t * Ast.Inner.typ_scheme * Rel.t) list

    val pp:
      Format.formatter ->
      (Ast.Var.t * Ast.Inner.typ_scheme * Rel.t) list ->
      unit
  end

  let get_analysis_module (logs: (module Logs.S)) options =
    let module Log = (val logs) in
    let module Options = (val get_module options) in
    begin match options.analysis_kind with
      | AnalysisKind.IndependentAttribute ->
        (module AnalyzerRel.Make(Log)(Options) : ANALYZER)
    end
end

let ( >>= ) m f = match m with
  | Error _ as err -> err
  | Ok v -> f v

let return x = Ok x

type package =
  | Pack: 'res * (Format.formatter -> 'res -> unit) -> package

type _ output_choice =
  | Value: package output_choice
  | Print: unit output_choice

let run (type a res) (out: res output_choice)
    (pp: Format.formatter -> a -> unit) : _ -> res = function
  | Error (loc, msg) -> Format.printf "%a@.%s@." Location.pp loc msg; exit 2
  | Ok v -> begin match out with
      | Print ->
        Format.printf "@[%a@]@." pp v
      | Value ->
        Pack (v, pp)
    end

let simpl_gen out options file =
  if options.Options.run_internal_tests
  then Tests.run ();
  let module Options = struct
    open Options
    let extensional_eq = false
    let extensional_bot = false
    let learn_from_patterns = options.learn_from_patterns
    let strict_errors = options.strict_errors
    let unroll_fixpoints = options.unroll_fixpoints
    let delay_widening = options.delay_widening
  end in
  let module Check = Typecheck.Make(Options) in
  let analyze pp interpret empty_env =
    run out pp @@ begin
      Driver.parse_file Parser.program file >>= fun prog ->
      Check.typecheck prog >>= fun prog' ->
      return @@ interpret empty_env prog'
    end
  in
  begin match options.analysis_kind with
    | IndependentAttribute ->
      let module Analyze = AnalyzerRel.Make(Logs.Std)(Options) in
      analyze Analyze.pp Analyze.interpret_program Analyze.Env.empty
  end

let simpl_print = simpl_gen Print

let simpl_value = simpl_gen Value
